#include <bitmapfonts.c>
#include <fonts/captionfont.dat.c>
//#include <muon1/particles.c> // Unused for now as Muon1's particles are not as diverse as PDG's
#include <muon1/constants.c>
#include <lset.c>
#include <rnd/fgetline.c>
#include <rnd/filestring.c>
#include <rnd/sgetline.c>

typedef struct
{
	int id; // PDG particle id I think
	float px,py,pz,p0, // 4-momentum in GeV/c
		mass, // Mass in GeV/c^2
		x,y,z,t; // 4-posiiton in fm
	float tform; // Formation time
} OscarPart;

typedef LSet Oscar,OscarFrame;

Oscar Oscar_load(const char *filename)
{ // Returns LSet(LSet(OscarPart))
	int n,m; unsigned len; LSet ret=LSet(OscarFrame);
	char *fd=filestring(filename,&len),*in=fd,l[300];
	if (!fd) return ret;
	for (n=3;n>0;n--) sgetline_noalloc(&in,l,300);
	int mallards; char str[100];
	while (sgetline_noalloc(&in,l,300))
	{
		if (sscanf(l," %d %d",&mallards,&m)!=2) break;
		LSet frame=LSet(OscarPart); OscarPart p;
		for (n=m;n>0;n--)
		{
			sgetline_noalloc(&in,l,300);
			sscanf(l," %d %d %f %f %f %f %f %f %f %f %f %f",&mallards,
				&p.id,&p.px,&p.py,&p.pz,&p.p0,&p.mass,&p.x,&p.y,&p.z,&p.t, // NB: use mass not sqrt(p0^2-pxyz^2) because of few digits
				&p.tform); 
			LSet_add(&frame,&p);
		}
		LSet_add(&ret,&frame);
		sprintf(str,"Loaded %d frames, %d bytes (%d%%)",ret.m,in-fd,(int)(100.0*(in-fd)/len));
		bfwrite((xres-bflength(str))/2,yres/2,str,GFX_white);
		ccl();
	}
	free(fd);
	return ret;
}

void Oscar_savebinary(const Oscar s,const char *filename)
{
	FILE *out=fopen(filename,"wb");
	fwrite(&s.m,1,sizeof(unsigned),out);
	OscarFrame *f=(OscarFrame *)s.a;
	for (int n=0;n<s.m;n++) LSet_write(&f[n],out);
	fclose(out);
}

Oscar Oscar_loadbinary(const char *filename)
{
	FILE *in=fopen(filename,"rb");
	int n,m;
	fread(&m,1,sizeof(unsigned),in);
	LSet ret=LSet(OscarFrame); OscarFrame f;
	for (n=0;n<m;n++)
	{
		f=LSet_read(in);
		LSet_add(&ret,&f);
	}
	fclose(in);
	return ret;
}

#include <rnd/cachefilename.c>
LOADCACHED(Oscar,load,loadbinary,savebinary)

void Oscar_free(Oscar *ps)
{
	LSet *a=(LSet *)ps->a;
	for (int n=ps->m-1;n>=0;n--) LSet_free(&a[n]);
	LSet_free(ps);
}

void colourless_sphere(const double ox,const double oy,const double oz,const double r,const int m=10)
{ // 4*m*m triangles // glBegin(GL_QUADS); ...
	int n,i; double *c=(double *)malloc((m+1)*sizeof(double)),*s=(double *)malloc((m+1)*sizeof(double)),th;
	for (n=m;n>=0;n--) {th=M_PI*n/m; c[n]=cos(th); s[n]=sin(th);}
	for (n=m-1;n>=0;n--) for (i=m-1;i>=0;i--)
	{
		glNormal3d(c[n]*s[i],s[n]*s[i],c[i]);
		glVertex3d(ox+r*c[n]*s[i],oy+r*s[n]*s[i],oz+r*c[i]);
		glNormal3d(c[n+1]*s[i],s[n+1]*s[i],c[i]);
		glVertex3d(ox+r*c[n+1]*s[i],oy+r*s[n+1]*s[i],oz+r*c[i]);
		glNormal3d(c[n+1]*s[i+1],s[n+1]*s[i+1],c[i+1]);
		glVertex3d(ox+r*c[n+1]*s[i+1],oy+r*s[n+1]*s[i+1],oz+r*c[i+1]);
		glNormal3d(c[n]*s[i+1],s[n]*s[i+1],c[i+1]);
		glVertex3d(ox+r*c[n]*s[i+1],oy+r*s[n]*s[i+1],oz+r*c[i+1]);
		glNormal3d(c[n]*s[i],-s[n]*s[i],c[i]);
		glVertex3d(ox+r*c[n]*s[i],oy-r*s[n]*s[i],oz+r*c[i]);
		glNormal3d(c[n+1]*s[i],-s[n+1]*s[i],c[i]);
		glVertex3d(ox+r*c[n+1]*s[i],oy-r*s[n+1]*s[i],oz+r*c[i]);
		glNormal3d(c[n+1]*s[i+1],-s[n+1]*s[i+1],c[i+1]);
		glVertex3d(ox+r*c[n+1]*s[i+1],oy-r*s[n+1]*s[i+1],oz+r*c[i+1]);
		glNormal3d(c[n]*s[i+1],-s[n]*s[i+1],c[i+1]);
		glVertex3d(ox+r*c[n]*s[i+1],oy-r*s[n]*s[i+1],oz+r*c[i+1]);
	}
	free(c); free(s);
}

void colourless_octohedron(const double ox,const double oy,const double oz,const double r)
{ // glBegin(GL_TRIANGLES); ...
	int x,y,z;
	for (x=-1;x<=1;x++) for (y=-1;y<=1;y++) for (z=-1;z<=1;z++)
	{
		glNormal3d(x/sqrt(3),y/sqrt(3),z/sqrt(3));
		glVertex3d(x*r,0,0);
		glVertex3d(0,y*r,0);
		glVertex3d(0,0,z*r);
	}
}

void OscarFrame_plotv1(OscarFrame f)
{ // Old version as far as anti-n,p
	//bfwrite(xres/2,yres/2,"Test",GFX_white);
	static GLuint spherelist=0,smallsph=0;
	if (!spherelist)
	{
		spherelist=glGenLists(1);
		glNewList(spherelist,GL_COMPILE);
		colourless_sphere(0,0,0,0.5);
		glEndList();
		smallsph=glGenLists(1);
		glNewList(smallsph,GL_COMPILE);
		colourless_sphere(0,0,0,0.1);
		glEndList();
	}
	glEnable(GL_LIGHTING); glEnable(GL_LIGHT0); glEnable(GL_COLOR_MATERIAL);
	OscarPart *p=(OscarPart *)f.a;
	int n; unsigned c; float r;
	for (n=f.m-1;n>=0;n--)
	{
		glPushMatrix(); glTranslatef(p[n].x,p[n].y,p[n].z);
		if (p[n].id==2212) {glColor3f(1,0.3,0.3); r=0.5;}
		else if (p[n].id==2112) {glColor3f(0.8,0.8,0.8); r=0.5;}
		else if (p[n].id==-2212) {glColor3f(0,0.7,0.8); r=0.5;}
		else if (p[n].id==-2112) {glColor3f(0.3,0.5,0.3); r=0.5;}
		else {glColor3f(0,1,0.5); r=0.1;}
		glCallList(r>0.3?spherelist:smallsph);
		glPopMatrix();
	}
	glDisable(GL_LIGHTING);
}

void gammadistort(const OscarPart *p,GLfloat *m)
{
	GLfloat k=(p->mass/p->p0-1)/(p->px*p->px+p->py*p->py+p->pz*p->pz);
	m[0]=1.0+k*p->px*p->px;
	m[1]=m[4]=k*p->px*p->py;
	m[2]=m[8]=k*p->px*p->pz;
	m[5]=1.0+k*p->py*p->py;
	m[6]=m[9]=k*p->py*p->pz;
	m[10]=1.0+k*p->pz*p->pz;
	glMultMatrixf(m);
}

const double r_proton=0.8775, // RMS charge radius in fm (OSCAR units)
	m_proton=0.938272046; // In GeV/c^2 (OSCAR units)

GLuint spherelist=0,antispherelist=0;

void init_lists(void)
{
	spherelist=glGenLists(21);
	for (int n=1;n<=20;n++)
	{
		glNewList(spherelist+n,GL_COMPILE);
		glBegin(GL_QUADS);
		colourless_sphere(0,0,0,1,n);
		glEnd();
		glEndList();
	}
	antispherelist=glGenLists(1);
	glNewList(antispherelist,GL_COMPILE);
	glBegin(GL_TRIANGLES);
	colourless_octohedron(0,0,0,sqrt(M_PI/2));
	glEnd();
	glEndList();
	/*sphereq=glGenLists(1);
	glNewList(sphereq,GL_COMPILE);
	colourless_sphere(0,0,0,1,4);
	glEndList();*/ // Tried: changing to sphereq doesn't improve performance
}

void plotquark(const float x,const float y,const float z,const int q,const float rot=0)
{
	float r;
	switch (abs(q))
	{
		case 1: glColor3f(0.6,0.6,0.4); r=0.1; break;
		case 2: glColor3f(0.8,0.6,0.4); r=0.1; break;
		case 3: glColor3f(0.4,0.4,0.8); r=0.15; break;
		case 4: glColor3f(0.4,0.8,0.4); r=0.2; break;
		case 5: glColor3f(0.6,0.4,0.8); r=0.25; break;
		case 6: glColor3f(0.8,0.6,0.8); r=0.3; break;
	} // Removing the translatef and push/popmatrix gives some speedup here
	glPushMatrix(); glTranslatef(x,y,z);
	glScalef(r,r,r);
	if (rot!=0) glRotatef(fmod(rot+90,90)-90,0,1,0);
	glCallList(q<0?antispherelist:spherelist+5);
	//glCallList(KEY(VK_SPACE)?sphereq:spherelist+10);
	glPopMatrix();
}

void OscarFrame_plot(OscarFrame f,const int showquarks=1)
{ // showquarks=1 means make formed hadrons transparent and show bound quarks
	int n,i,j;
	if (!spherelist) init_lists();
	static float sa[90]; static int sa_init=0;
	if (!sa_init)
	{
		for (n=89;n>=0;n--) sa[n]=sin(M_TWOPI*n/90);
		sa_init=1;
	}
	glEnable(GL_LIGHTING); glEnable(GL_LIGHT0); glEnable(GL_COLOR_MATERIAL);
	glEnable(GL_NORMALIZE); // Because we are scaling spheres with normals
	OscarPart *p=(OscarPart *)f.a;
	int charge,q[3],mq; float r,th,dim=showquarks?0.2:1;
	GLfloat m[16]; m[3]=m[7]=m[11]=m[12]=m[13]=m[14]=0; m[15]=1;
	for (n=f.m-1;n>=0;n--) if (showquarks || p[n].t<p[n].tform) if (p[n].id)
	{
		i=abs(p[n].id);
		q[0]=i/10%10; q[1]=i/100%10; q[2]=i/1000%10; // Quarks usually in ascending order here
		mq=q[2]?3:2;
		if (mq==3) {if (p[n].id<0) {q[0]*=-1; q[1]*=-1; q[2]*=-1;}} // Baryon {anti-baryon}
		else // Meson
		{
			j=(p[n].id>0)^(q[1]%2); 
			if (j) q[0]*=-1; else q[1]*=-1;
		}
		glPushMatrix(); glTranslatef(p[n].x,p[n].y,p[n].z);
		gammadistort(&p[n],m);
		i=40*p[n].t*p[n].mass/p[n].p0;
		for (j=mq-1;j>=0;j--,i+=90/mq)
			plotquark(0.25*sa[i%90],0.25*sa[(i+30)%90],0.25*sa[(i+60)%90],q[j]);
		glPopMatrix();
	}
	if (showquarks) {glEnable(GL_BLEND); glBlendFunc(GL_ONE,GL_ONE_MINUS_SRC_COLOR); glDepthMask(0);}
	for (n=f.m-1;n>=0;n--) if (p[n].t>=p[n].tform)
	{
		i=abs(p[n].id);
		q[0]=i/10%10; q[1]=i/100%10; q[2]=i/1000%10; // Quarks usually in ascending order here
		if (q[2]) charge=2-q[0]%2-q[1]%2-q[2]%2; // Baryon
		else charge=(q[0]%2!=q[1]%2); // For PDG, positive mesons are the "particles"
		if (p[n].id<0) charge*=-1;
		switch (charge)
		{
			case -2: glColor3f(0,0,0.8*dim); break;
			case -1: glColor3f(0.4*dim,0.7*dim,0.8*dim); break;
			case 0: glColor3f(0.8*dim,0.8*dim,0.8*dim); break;
			case 1: glColor3f(0.8*dim,0.4*dim,0.4*dim); break;
			case 2: glColor3f(0.8*dim,0,0); break;
		}
		r=pow(p[n].mass/m_proton,1.0/3)*r_proton;
		glPushMatrix(); glTranslatef(p[n].x,p[n].y,p[n].z);
		gammadistort(&p[n],m);
		glScalef(r,r,r);
		glCallList(spherelist+10);
		glPopMatrix();
	}
	if (showquarks) {glDisable(GL_BLEND); glDepthMask(1);}
	glDisable(GL_LIGHTING);
	glDisable(GL_NORMALIZE);
}

void init_lists_v2(void)
{
	spherelist=glGenLists(21);
	for (int n=1;n<=20;n++)
	{
		glNewList(spherelist+n,GL_COMPILE);
		colourless_sphere(0,0,0,1,n);
		glEndList();
	}
	antispherelist=glGenLists(1);
	glNewList(antispherelist,GL_COMPILE);
	colourless_octohedron(0,0,0,sqrt(M_PI/2));
	glEndList();
}

char *vertprog_v2=(char *)"#version 120\n\
varying vec3 pos,norm;\n\
void main()\n\
{\n\
	pos=gl_MultiTexCoord0.xyz+gl_MultiTexCoord0.w*gl_Vertex.xyz;\n\
	norm=gl_Normal.xyz;\n\
	gl_Position=gl_ModelViewProjectionMatrix*vec4(pos,1);\n\
	gl_FrontColor=gl_Color;\n\
}";

char *fragprog_v2=(char *)"\n\
varying vec3 pos,norm;\n\
void main()\n\
{\n\
	gl_FragColor=gl_Color*vec4(vec3(dot(norm,vec3(0,0,1))),1);\n\
}";

GLuint shaderprog_v2(void)
{
	static GLuint ret=0;
	if (!ret) ret=shaderprogvf(vertprog_v2,fragprog_v2);
	return ret;
}

void plotquark_v2(const float x,const float y,const float z,const int q,const float rot=0)
{
	float r;
	switch (abs(q))
	{
		case 1: glColor3f(0.6,0.6,0.4); r=0.1; break;
		case 2: glColor3f(0.8,0.6,0.4); r=0.1; break;
		case 3: glColor3f(0.4,0.4,0.8); r=0.15; break;
		case 4: glColor3f(0.4,0.8,0.4); r=0.2; break;
		case 5: glColor3f(0.6,0.4,0.8); r=0.25; break;
		case 6: glColor3f(0.8,0.6,0.8); r=0.3; break;
	} // Removing the translatef and push/popmatrix gives some speedup here
	glTexCoord4f(x,y,z,r);
	//if (rot!=0) glRotatef(fmod(rot+90,90)-90,0,1,0);
	glCallList(q<0?antispherelist:spherelist+3);
	//glCallList(KEY(VK_SPACE)?sphereq:spherelist+10);
}

void OscarFrame_plot_v2(OscarFrame f,const int showquarks=1)
{ // showquarks=1 means make formed hadrons transparent and show bound quarks
	int n,i,j;
	if (!spherelist) init_lists_v2();
	static float sa[90]; static int sa_init=0;
	if (!sa_init)
	{
		for (n=89;n>=0;n--) sa[n]=sin(M_TWOPI*n/90);
		sa_init=1;
	}
	OscarPart *p=(OscarPart *)f.a;
	int charge,q[3],mq; float r,th,dim=showquarks?0.2:1;
	GLfloat m[16]; m[3]=m[7]=m[11]=m[12]=m[13]=m[14]=0; m[15]=1;
	if (!hasshaderprocs()) reportfatal("Shaders unavailable");
	glUseProgram(shaderprog_v2());
	glBegin(GL_TRIANGLES);
	for (n=f.m-1;n>=0;n--) if (showquarks || p[n].t<p[n].tform) if (p[n].id)
	{
		i=abs(p[n].id);
		q[0]=i/10%10; q[1]=i/100%10; q[2]=i/1000%10; // Quarks usually in ascending order here
		mq=q[2]?3:2;
		if (mq==3) {if (p[n].id<0) {q[0]*=-1; q[1]*=-1; q[2]*=-1;}} // Baryon {anti-baryon}
		else // Meson
		{
			j=(p[n].id>0)^(q[1]%2); 
			if (j) q[0]*=-1; else q[1]*=-1;
		}
		//gammadistort(&p[n],m);
		i=40*p[n].t*p[n].mass/p[n].p0;
		for (j=mq-1;j>=0;j--,i+=90/mq)
			plotquark_v2(p[n].x+0.25*sa[i%90],p[n].y+0.25*sa[(i+30)%90],p[n].z+0.25*sa[(i+60)%90],q[j]);
	}
	glEnd();
	if (showquarks) {glEnable(GL_BLEND); glBlendFunc(GL_ONE,GL_ONE_MINUS_SRC_COLOR); glDepthMask(0);}
	glBegin(GL_QUADS);
	for (n=f.m-1;n>=0;n--) if (p[n].t>=p[n].tform)
	{
		i=abs(p[n].id);
		q[0]=i/10%10; q[1]=i/100%10; q[2]=i/1000%10; // Quarks usually in ascending order here
		if (q[2]) charge=2-q[0]%2-q[1]%2-q[2]%2; // Baryon
		else charge=(q[0]%2!=q[1]%2); // For PDG, positive mesons are the "particles"
		if (p[n].id<0) charge*=-1;
		switch (charge)
		{
			case -2: glColor3f(0,0,0.8*dim); break;
			case -1: glColor3f(0.4*dim,0.7*dim,0.8*dim); break;
			case 0: glColor3f(0.8*dim,0.8*dim,0.8*dim); break;
			case 1: glColor3f(0.8*dim,0.4*dim,0.4*dim); break;
			case 2: glColor3f(0.8*dim,0,0); break;
		}
		r=pow(p[n].mass/m_proton,1.0/3)*r_proton;
		glTexCoord4f(p[n].x,p[n].y,p[n].z,r);
		//gammadistort(&p[n],m);
		glCallList(spherelist+5);
	}
	glEnd();
	glUseProgram(0);
	if (showquarks) {glDisable(GL_BLEND); glDepthMask(1);}
}

void plotting_benchmark(const char *filename)
{
	FILE *out=fopen(filename,"wt");
	
	fclose(out);
	exit(0);
}

void fontsize(const float s)
{ // DPI-aware font size (s=1=normal)
	char str[100]; sprintf(str,":scale=%g",s*Maxi(1,(yres+500)/1000)); bfsetfont(str);
}

int touch=0;

void plotkey(const float ox,float oy)
{
	fontsize(0.5);
	bfwrite(ox,oy,"Key:",GFX_white); oy+=BF_height*2;
	int n; float tab=BF_height*5,py;
	bfwrite(ox,oy,"Quarks",GFX_white); bfwrite(ox+tab,oy,"Anti-quarks",GFX_white);
	tab*=1.5;
	const char *qname[]={"0","down","up","strange","charm"};
	for (n=1;n<=4;n++)
	{
		py=oy+n*BF_height*2;
		bfwrite(ox+0.5*tab+BF_height-0.5*bflength(qname[n]),py,qname[n],GFX_white);
	}
	ortho_on();
	glPushMatrix();
	float s=BF_height*5,rot=eithertime()*50;
	glScalef(s,s,s);
	glEnable(GL_LIGHTING); glEnable(GL_NORMALIZE);
	for (n=1;n<=4;n++)
	{
		py=oy+n*BF_height*2;
		plotquark((ox+BF_height)/s,(py+0.5*BF_height)/s,0,n,rot);
		plotquark((ox+tab+BF_height)/s,(py+0.5*BF_height)/s,0,-n,rot);
	}
	glPopMatrix();
	oy+=BF_height*10;
	bfwrite(ox,oy,"Nucleons",GFX_white);
	const char *cname[]={"negative","neutral","positive"};
	oy+=BF_height*4;
	for (n=-1;n<=1;n++)
	{
		py=oy+n*BF_height*2;
		bfwrite(ox+0.25*tab+BF_height,py,cname[1+n],GFX_white);
	}
	s=BF_height;
	glPushMatrix();
	glScalef(s,s,s);
	glEnable(GL_BLEND); glBlendFunc(GL_ONE,GL_ONE_MINUS_SRC_COLOR); glDepthMask(0);
	float dim=0.2;
	for (n=-1;n<=1;n++)
	{
		py=oy+n*BF_height*2;
		switch (n)
		{
			case -2: glColor3f(0,0,0.8*dim); break;
			case -1: glColor3f(0.4*dim,0.7*dim,0.8*dim); break;
			case 0: glColor3f(0.8*dim,0.8*dim,0.8*dim); break;
			case 1: glColor3f(0.8*dim,0.4*dim,0.4*dim); break;
			case 2: glColor3f(0.8*dim,0,0); break;
		}
		glPushMatrix(); glTranslatef((ox+BF_height)/s,(py+0.5*BF_height)/s,0);
		glScalef(r_proton,r_proton,r_proton);
		glCallList(spherelist+10);
		glPopMatrix();
	}
	glDisable(GL_BLEND); glDepthMask(1);
	glDisable(GL_LIGHTING); glDisable(GL_NORMALIZE);
	glPopMatrix();
	oy+=BF_height*5;
	bfwrite(ox,oy,"Touchscreen display:\n\nDrag to rotate\nPinch to zoom\nTime controls below",0x8080C0);
	ortho_off();
	fontsize(1);
}

#include <plot3D.c>

OscarFrame curframe; int showquarks=1;

void plotfn(void *nullards)
{
	static GLuint list=0; // No speed-up from making it a list
	if (!list && KEY(VK_L))
	{
		list=glGenLists(1);
		glNewList(list,GL_COMPILE);
		OscarFrame_plot(curframe,showquarks);
		glEndList();
	}
	if (list) glCallList(list);
	else if (KEY(VK_SHIFT)) OscarFrame_plot_v2(curframe,showquarks);
	else OscarFrame_plot(curframe,showquarks);
}

void Oscar_particleinventory(const Oscar s,const char *filename)
{
	int n,i,j;
	typedef struct {int id,n; double mass,pmass;} PIDcount;
	LSet cs=LSet(PIDcount);
	OscarFrame *f=(OscarFrame *)s.a; OscarPart *p; PIDcount *c;
	for (n=s.m-1;n>=0;n--)
	{
		p=(OscarPart *)f[n].a;
		for (i=f[n].m-1;i>=0;i--)
		{
			c=(PIDcount *)cs.a;
			for (j=cs.m-1;j>=0;j--) if (p[i].id==c[j].id)
				{c[j].n++; j=-10; break;}
			if (j>-10)
			{
				PIDcount nc; nc.id=p[i].id; nc.n=1;
				nc.mass=p[i].mass;
				nc.pmass=sqrt(sq(p[i].p0)-sq(p[i].px)-sq(p[i].py)-sq(p[i].pz));
				LSet_add(&cs,&nc);
			}
		}
	}
	FILE *out=fopen(filename,"wt");
	c=(PIDcount *)cs.a;
	fprintf(out,"id,n,mass(first),pmass(first)\n");
	for (n=cs.m-1;n>=0;n--) fprintf(out,"%d,%d,%.15lg,%.15lg\n",c[n].id,c[n].n,c[n].mass,c[n].pmass);
	fclose(out);
	LSet_free(&cs);
}

void adddigitspacers(char *s0,const int p=6)
{ // Warning: can overflow allocation
	char *s=strchr(s0,'.'),*t;
	if (!s) return;
	for (int d=0;*s;s++) if (isdigit(*s))
	{
		d++;
		if (d%p==0)
		{
			for (t=s+strlen(s);t>s;t--) t[1]=*t;
			s++; *s=' '; 
		}
	}
}

#include <rnd/setcwdtohomedir.c>
#include <rnd/istouchscreen.c>
#include <rgbops.c>

int touchicon(const float x,const float y,const float s,const int type,unsigned c)
{
	float u=s*0.1; int on=(mx>=x && mx<x+s && my>=y && my<y+s);
	if (!on) c=and24(c,0x808080);
	switch (type)
	{
		case 0: // Pause
			rect(x+u,y+u,u*3,u*8,c);
			rect(x+u*5,y+u,u*3,u*8,c);
			break;
		case 1: // Play
			glColor3ub(c>>16,(c>>8)&0xFF,c&0xFF);
			glBegin(GL_TRIANGLES);
			glVertex2f(x+u,y+u);
			glVertex2f(x+u,y+u*9);
			glVertex2f(x+u*9,y+u*5);
			glEnd();
			break;
		case 2: // Back
			rect(x+u,y+u,u,u*8,c);
			glColor3ub(c>>16,(c>>8)&0xFF,c&0xFF);
			glBegin(GL_TRIANGLES);
			glVertex2f(x+u*9,y+u);
			glVertex2f(x+u*9,y+u*9);
			glVertex2f(x+u*2,y+u*5);
			glEnd();
			break;
		case 3: // Forward
			rect(x+u*8,y+u,u,u*8,c);
			glColor3ub(c>>16,(c>>8)&0xFF,c&0xFF);
			glBegin(GL_TRIANGLES);
			glVertex2f(x+u,y+u);
			glVertex2f(x+u,y+u*9);
			glVertex2f(x+u*8,y+u*5);
			glEnd();
			break;
	}
	return on;
}

LSet conf;
Oscar fs; int sim,frame,paused; float dist;
int omx,omy;

void advancesim(const int dsim)
{
	Oscar_free(&fs);
	sim+=dsim; frame=0;
	char *file=configgetnth(&conf,"filename",sim);
	if (!file)
	{
		if (configgetb(&conf,"loop")) {if (sim<1) sim=3; else sim=1; file=configgetnth(&conf,"filename",sim);}
		else exit(0);
	}
	fs=Oscar_loadcached(file);
}

int touchbar(const float ox,const float oy,const float xs,const float ys)
{ // Returns if clicked on
	ortho_on();
	float g=Min(0.1*yres,Min(xs,ys)); int ret=0,moved=(mx!=omx || my!=omy);
	if (touchicon(ox,oy,g,paused?1:0,0x00FF00)) {if (moved) paused^=1; ret=1;}
	if (touchicon(ox+xs-g*2,oy,g,2,0xFFC000)) {if (moved) advancesim(-1); ret=1;}
	if (touchicon(ox+xs-g,oy,g,3,0xFFC000)) {if (moved) advancesim(1); ret=1;}
	float bx=ox+g*1.5,bxs=xs-g*4,by=oy+0.2*ys,bys=ys*0.6;
	if (mx>=bx && mx<bx+bxs && my>=oy && my<oy+ys)
	{
		if (moved) frame=nonegi(Mini(fs.m-1,(float)(mx-bx)/bxs*fs.m));
		ret=1;
	}
	rect(bx,by,bxs,bys,0x404080);
	rect(bx,by,bxs*(0.5+frame)/fs.m,bys,0x8080C0);
	ortho_off();
	return ret;
}

void oscar3d_gesture_handler(WPARAM wp,LPARAM lp)
{
	GESTUREINFO gi;
	ZeroMemory(&gi,sizeof(GESTUREINFO));
	gi.cbSize=sizeof(GESTUREINFO);
	if (!GetGestureInfo((HGESTUREINFO)lp,&gi)) return;
	POINT p; p.x=gi.ptsLocation.x; p.y=gi.ptsLocation.y;
	ScreenToClient(gi.hwndTarget,&p);
	static int zooming=0;
	if (gi.dwID==GID_END) zooming=0;
	else if (gi.dwID==GID_ZOOM)
	{
		float z=gi.ullArguments; static float initzoom=1;
		if (!zooming) initzoom=dist*z;
		dist=Min(100,Max(1,initzoom/z));
		//FILE *out=fopen("zoom.log","at");
		//fprintf(out,"z=%g zooming=%d initzoom=%g dist=%g\n",z,zooming,initzoom,dist);
		//fclose(out);
		zooming=1;
	}
	CloseGestureInfoHandle((HGESTUREINFO)lp);
}

WINMAIN
{ /**/ // Needs error message for file not found
	intel_hd_graphics_cludge=1;
	setcwdtohomedir();
	char str[100],*s;
	conf=configload("config.txt");
	int stereomode,swapstereo; plot3D_config(&conf,&stereomode,&swapstereo);
	GRAPHICS3D(xres,yres,"UrQMD OSCAR File Visualisation [Stephen Brooks 2014-19]");
	loadfontresource(captionfont_dat,"caption");
	fontsize(1);
	touch=istouchscreen();
	if (touch) {windows_touchscreen_init(); wm_gesture_handler=oscar3d_gesture_handler;}
	//Oscar_particleinventory(fs,"pids.csv");
	int stereo=configgetb(&conf,"Stereo"),showkey=1;
	sim=1; fs=Oscar_loadcached(configgetnth(&conf,"filename",sim));
	frame=0; paused=0; dist=18;
	float defang=configgetd(&conf,"Default angle",90),rot=180.0-defang,vrot=0;
	double tframe=0;
	omx=mx; omy=my;
	LOOP
	{
		double t0=eithertime();
		if (KEY(VK_Z) || !touch && (mb&1)) dist=Max(0.95*dist,1);
		if (KEY(VK_X) || !touch && (mb&2)) dist=Min(1.05*dist,100);
		if (KHIT(VK_2)) stereo=0;
		if (KHIT(VK_3)) {if (KEY(VK_SHIFT)) swapstereo^=1; else stereo=1;}
		if (KHIT(VK_P)) paused^=1;
		if (KHIT(VK_R)) frame=0;
		if (KHIT(VK_T)) frame=fs.m-1;
		if (KHIT(VK_A)) advancesim(-1);
		if (!KEY(VK_CONTROL) && KHIT(VK_S)) advancesim(1);
		if (KEY(VK_LEFT)) frame=noneg(frame-1);
		if (KEY(VK_RIGHT)) frame=Mini(frame+1,fs.m-1);
		if (KHIT(VK_B)) showquarks^=1;
		if (KHIT(VK_K)) showkey^=1;
		glTranslatef(0,0,-dist);
		if (touch)
		{
			if (!touchbar(0,yres*0.9,xres,Min(xres,yres)*0.1))
				if (omb&mb&1) {rot+=360.0*(mx-omx)/xres; vrot+=180.0*(my-omy)/yres;}
		}
		else
		{
			rot=360.0*mx/xres-defang;
			vrot=180.0*my/yres-90.0;
		}
		glRotatef(rot,0,1,0);
		glRotatef(vrot,0,0,1);
		curframe=((OscarFrame *)fs.a)[frame];
		plot3D(plotfn,NULL,stereomode,swapstereo,0.5,-0.05);
		OscarPart *p=(OscarPart *)curframe.a;
		double t=p[0].t*1e-15/C;
		sprintf(str,"t = %.24lf s",t); adddigitspacers(str);
		bfwrite(0,0,str,GFX_white);
		sprintf(str,"Ion Collision");
		bfwrite((xres-bflength(str))/2,0,str,GFX_white);
		if (s=configgetnth(&conf,"title",sim)) bfwrite(xres-bflength(s)-BF_height,0,s,GFX_white);
		if (s=configget(&conf,"credit")) bfwrite(0,yres-BF_height-(touch?yres*0.1:0),s,GFX_white);
		if (showkey) plotkey(xres-BF_height*5.5,BF_height*2);
		if (!paused) frame++;
		if (frame>=fs.m) advancesim(1);
		if (KEY(VK_CONTROL) && KHIT(VK_S)) screenshotauto("UrQMD");
sprintf(str,"%lg",tframe); bfwrite(0,BF_height*2,str,GFX_white);
		omx=mx; omy=my;
		ccl();
		tframe=eithertime()-t0;
		while (eithertime()<t0+1.0/60);
	}
	Oscar_free(&fs);
	configsave(&conf,"config.txt");
	configfree(&conf);
}
